@@ -1,9 +1,13 @@
 from __future__ import print_function
 from __future__ import absolute_import
+from operator import itemgetter
 from builtins import str
 import pandas as pd
 from builtins import object
 import warnings
+from autoprotocol import Unit
+from requests.exceptions import ReadTimeout
+from copy import deepcopy
 
 
 def _check_api(obj_type):
@@ -169,6 +173,8 @@ class Run(_BaseObject):
             myRun = Run('r12345')
             myRun.data
             myRun.instructions
+            myRun.containers
+            myRun.Instructions[0]
 
     Attributes
     ----------
@@ -177,9 +183,11 @@ class Run(_BaseObject):
     name: str
         Run name
     data: DataFrame
-        DataFrame of all datasets which belong to this run
+        DataFrame summary of all datasets which belong to this run
     instructions: DataFrame
-        DataFrame of all Instruction objects which belong to this run
+        DataFrame summary of all Instruction objects which belong to this run
+    containers: DataFrame
+        DataFrame summary of all Container objects which belong to this run
     project_id : str
         Project id which run belongs to
     attributes: dict
@@ -244,6 +252,19 @@ class Run(_BaseObject):
             self._instructions.insert(2, "Completed", [inst.completed_at for inst in self._instructions.Instructions])
         return self._instructions
 
+    @property
+    def Instructions(self):
+        """
+        Helper for allowing direct access of `Instruction` objects
+
+        Returns
+        -------
+        Series
+            Returns a Series of `Instruction` objects
+
+        """
+        return self.instructions.Instructions
+
     @property
     def containers(self):
         if self._containers.empty:
@@ -253,12 +274,24 @@ class Run(_BaseObject):
             self._containers = pd.DataFrame(container_list)
             self._containers.columns = ["Containers"]
             self._containers.insert(0, "Name", [container.name for container in self._containers.Containers])
-            self._containers.insert(1, "container_id", [container.id for container in self._containers.Containers])
+            self._containers.insert(1, "ContainerId", [container.id for container in self._containers.Containers])
             self._containers.insert(2, "Type", [container.container_type.shortname for container in self._containers.Containers])
             self._containers.insert(3, "Status", [container.attributes["status"] for container in self._containers.Containers])
             self._containers.insert(4, "Storage Condition", [container.storage for container in self._containers.Containers])
         return self._containers
 
+    @property
+    def Containers(self):
+        """
+        Helper for allowing direct access of `Container` objects
+
+        Returns
+        -------
+        Series
+            Returns a Series of `Container` objects
+        """
+        return self.containers.Containers
+
     @property
     def data(self):
         """
@@ -284,11 +317,27 @@ class Run(_BaseObject):
                     self._data = pd.DataFrame(sorted(list(data_dict.items()), key=lambda x: x[0]))
                     self._data.columns = ["Name", "Datasets"]
                     self._data.insert(1, "DataType", ([ds.operation for ds in self._data.Datasets]))
-                except Exception:
+                except ReadTimeout:
                     print('Operation timed out after %d seconds. Returning data_ids instead of Datasets.\nTo try again, increase value of self.timeout and resubmit request.' % self.timeout)
                     return self.data_ids
         return self._data
 
+    @property
+    def Datasets(self):
+        """
+        Helper for allowing direct access of `Dataset` objects
+
+        Returns
+        -------
+        Series
+            Returns a Series of `Dataset` objects
+        """
+        try:
+            return self.data.Datasets
+        except Exception:
+            print('Unable to load Datasets successfully. Returning empty series.')
+            return pd.Series()
+
     def _repr_html_(self):
         return """<iframe src="%s" frameborder="0" allowtransparency="true" \
         style="height:450px" seamless></iframe>""" % \
@@ -340,24 +389,64 @@ class Dataset(_BaseObject):
         self.id = data_id
         self.operation = self.attributes["instruction"]["operation"]["op"]
         self.data_type = self.attributes["data_type"]
+        self._raw_data = None
         self._data = pd.DataFrame()
         self.container = Container(self.attributes["container"]["id"], attributes=self.attributes["container"],
                                    connection=connection)
 
+    @property
+    def raw_data(self):
+        if not self._raw_data:
+            # Get all raw data
+            self._raw_data = self.connection.dataset(data_id=self.id, key="*")
+        return self._raw_data
+
     @property
     def data(self, key="*"):
         if self._data.empty:
             # Get all data initially (think about lazy loading in the future)
-            self._data = pd.DataFrame(self.connection.dataset(data_id=self.id, key="*"))
+            try:
+                self._data = pd.DataFrame(self.raw_data)
+            except:
+                raise RuntimeError("Failed to cast data as DataFrame. Try using raw_data property instead.")
             self._data.columns = [x.upper() for x in self._data.columns]
         if key == "*":
             return self._data
         else:
             return self._data[key]
 
+    def cross_ref_aliquots(self):
+        # Use the container.aliquots DataFrame as the base
+        aliquot_data = deepcopy(self.container.aliquots)
+        data_column = []
+        indices_without_data = []
+        # Print a warning if new column will overwrite existing column
+        if "Aliquot Data" in aliquot_data.columns.values.tolist():
+            warnings.warn("Column 'Aliquot Data' will be overwritten with data pulled from Dataset.")
+        # Look up data for every well index
+        for index in aliquot_data.index:
+            # Get humanized index
+            humanized_index = self.container.container_type.humanize(int(index))
+            if humanized_index in self.data:
+                # Use humanized index to get data for that well
+                data_point = self.data.loc[0, humanized_index]
+            else:
+                # If no data for that well, use None instead
+                data_point = None
+                indices_without_data.append(humanized_index)
+            # Append data point to list
+            data_column.append(data_point)
+        # Print a list of well indices that do not have corresponding data keys
+        if len(indices_without_data) > 0:
+            warnings.warn("The following indices were not found as data keys: %s" % ", ".join(indices_without_data))
+        # Add these data as a column to the DataFrame
+        aliquot_data["Aliquot Data"] = data_column
+
+        return aliquot_data
+
     def _repr_html_(self):
         return """<iframe src="%s" frameborder="0" allowtransparency="true" \
-            style="height:500px" seamless></iframe>""" % \
+            style="height:400px; width:450px" seamless></iframe>""" % \
                self.connection.get_route('view_data', data_id=self.id)
 
 
@@ -378,8 +467,8 @@ class Instruction(object):
             myRun.instructions
 
             # Access instruction object
-            myRun.instructions.Instructions[1]
-            myRun.instructions.Instructions[1].warps
+            myRun.Instructions[1]
+            myRun.Instructions[1].warps
 
 
     Attributes
@@ -432,10 +521,19 @@ class Instruction(object):
     def warps(self):
         if self._warps.empty:
             warp_list = self.attributes["warps"]
-            self._warps = pd.DataFrame(x['command'] for x in warp_list)
-            self._warps.columns = [x.title() for x in self._warps.columns.tolist()]
-            self._warps.insert(1, "Started", [x["reported_started_at"] for x in warp_list])
-            self._warps.insert(2, "Completed", [x["reported_completed_at"] for x in warp_list])
+            if len(warp_list) != 0:
+                self._warps = pd.DataFrame(x['command'] for x in warp_list)
+                self._warps.columns = [x.title() for x in self._warps.columns.tolist()]
+                # Rearrange columns to start with `Name`
+                if "Name" in self._warps.columns:
+                    col_names = ["Name"] + [col for col in self._warps.columns if col != "Name"]
+                    self._warps = self._warps[col_names]
+                self._warps.insert(1, "WarpId", [x["id"] for x in warp_list])
+                self._warps.insert(2, "Completed", [x["reported_completed_at"] for x in warp_list])
+                self._warps.insert(3, "Started", [x["reported_started_at"] for x in warp_list])
+            else:
+                warnings.warn("There are no warps associated with this instruction. Please contact "
+                              "Transcriptic for assistance.")
         return self._warps
 
     def monitoring(self, data_type='pressure', grouping=None):
@@ -463,7 +561,7 @@ class Instruction(object):
 
     def _repr_html_(self):
         return """<iframe src="%s" frameborder="0" allowtransparency="true" \
-            style="height:1000px" seamless></iframe>""" % \
+            style="width:450px" seamless></iframe>""" % \
                self.connection.get_route('view_instruction', run_id=self.attributes["run_id"],
                                          project_id=self.attributes["project_id"], instruction_id=self.id)
 
@@ -492,7 +590,8 @@ class Container(_BaseObject):
     well_map: dict
         Well mapping with well indices for keys and well names as values
     aliquots: DataFrame
-        DataFrame of aliquots present in the container
+        DataFrame of aliquots present in the container. DataFrame index
+        now corresponds to the Well Index.
     container_type: autoprotocol.container_type.ContainerType
         Autoprotocol ContainerType object with many useful container type
         information and functions.
@@ -547,7 +646,7 @@ class Container(_BaseObject):
         container_type = self.attributes["container_type"]
 
         # Return the corresponding AP-Py container object for now. In the future, consider merging
-        # the current and future dictionary when instantianting container_type
+        # the current and future dictionary when instantiating container_type
         try:
             return _CONTAINER_TYPES[container_type["shortname"]]
         except KeyError:
@@ -556,11 +655,19 @@ class Container(_BaseObject):
 
     @property
     def aliquots(self):
+        """
+        Return a DataFrame of aliquots in the container, along with aliquot
+        name, volume, and properties. Row index for the DataFrame corresponds
+        to the well index of the aliquot.
+
+        """
         if self._aliquots.empty:
             aliquot_list = self.attributes["aliquots"]
-            self._aliquots = pd.DataFrame([dict({'Name': x['name'], 'Id': x['id'],
-                                                 'Volume': x['volume_ul']}, **x['properties'])
-                                           for x in aliquot_list])
+            self._aliquots = pd.DataFrame(sorted([dict({'Well Index': x['well_idx'], 'Name': x['name'], 'Id': x['id'],
+                                                 'Volume': Unit(float(x['volume_ul']), 'microliter')}, **x['properties'])
+                                           for x in aliquot_list], key=itemgetter('Well Index')))
+            indices = self._aliquots.pop('Well Index')
+            self._aliquots.set_index(indices, inplace=True)
         return self._aliquots
 
     def __repr__(self):
@@ -570,3 +677,4 @@ class Container(_BaseObject):
 
         """
         return "Container(%s)" % (str(self.name))
+
